﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using System.Collections.Generic;
using HeuristicLab.Problems.DataAnalysis.Evaluators;
using HeuristicLab.Analysis;
using HeuristicLab.Common;

namespace HeuristicLab.Problems.DataAnalysis.Operators {
  [Item("Covariant Parsimony Pressure", "Covariant Parsimony Pressure.")]
  [StorableClass]
  public class CovariantParsimonyPressure : SingleSuccessorOperator {
    public IScopeTreeLookupParameter<SymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<SymbolicExpressionTree>)Parameters["SymbolicExpressionTree"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> AdjustedQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["AdjustedQuality"]; }
    }
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IValueLookupParameter<DoubleValue> KParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["K"]; }
    }
    public ILookupParameter<DoubleValue> CParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["C"]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Generations"]; }
    }
    public IValueLookupParameter<IntValue> FirstGenerationParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["FirstGenerationParameter"]; }
    }
    public IValueLookupParameter<BoolValue> ApplyParsimonyPressureParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["ApplyParsimonyPressure"]; }
    }
    public ILookupParameter<DoubleValue> LengthCorrelationParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Correlation(Length, AdjustedFitness)"]; }
    }
    public ILookupParameter<DoubleValue> FitnessCorrelationParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Correlation(Fitness, AdjustedFitness)"]; }
    }
    public IValueLookupParameter<PercentValue> ComplexityAdaptionParameter {
      get { return (IValueLookupParameter<PercentValue>)Parameters["ComplexityAdaption"]; }
    }
    public IValueLookupParameter<BoolValue> InvertComplexityAdaptionParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["InvertComplexityAdaption"]; }
    }
    public IValueLookupParameter<DoubleValue> MinAverageSizeParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["MinAverageSize"]; }
    }

    protected CovariantParsimonyPressure(bool deserializing) : base(deserializing) { }
    protected CovariantParsimonyPressure(CovariantParsimonyPressure original, Cloner clone) : base(original, clone) { }
    public CovariantParsimonyPressure()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<SymbolicExpressionTree>("SymbolicExpressionTree"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("AdjustedQuality"));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization"));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("K", new DoubleValue(1.0)));
      Parameters.Add(new LookupParameter<IntValue>("Generations"));
      Parameters.Add(new ValueLookupParameter<IntValue>("FirstGenerationParameter", new IntValue(1)));
      Parameters.Add(new ValueLookupParameter<BoolValue>("ApplyParsimonyPressure"));
      Parameters.Add(new ValueLookupParameter<PercentValue>("ComplexityAdaption", new PercentValue(-0.01)));
      Parameters.Add(new LookupParameter<DoubleValue>("Correlation(Length, AdjustedFitness)"));
      Parameters.Add(new LookupParameter<DoubleValue>("Correlation(Fitness, AdjustedFitness)"));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("MinAverageSize", new DoubleValue(15)));
      Parameters.Add(new LookupParameter<DoubleValue>("C"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("InvertComplexityAdaption"));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CovariantParsimonyPressure(this, cloner);
    }


    [StorableHook(Persistence.Default.CompositeSerializers.Storable.HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey("Maximization"))
        Parameters.Add(new LookupParameter<BoolValue>("Maximization"));
      if (!Parameters.ContainsKey("K"))
        Parameters.Add(new ValueLookupParameter<DoubleValue>("K", new DoubleValue(1.0)));
      if (!Parameters.ContainsKey("AdjustedQuality")) {
        Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("AdjustedQuality"));
      }
      if (!Parameters.ContainsKey("Generations")) {
        Parameters.Add(new LookupParameter<IntValue>("Generations"));
      }
      if (!Parameters.ContainsKey("FirstGenerationParameter")) {
        Parameters.Add(new ValueLookupParameter<IntValue>("FirstGenerationParameter", new IntValue(1)));
      }
      if (!Parameters.ContainsKey("ApplyParsimonyPressure")) {
        Parameters.Add(new ValueLookupParameter<BoolValue>("ApplyParsimonyPressure"));
      }
      if (!Parameters.ContainsKey("ComplexityAdaption")) {
        Parameters.Add(new ValueLookupParameter<PercentValue>("ComplexityAdaption", new PercentValue(-0.01)));
      }
      if (!Parameters.ContainsKey("MinAverageSize")) {
        Parameters.Add(new ValueLookupParameter<DoubleValue>("MinAverageSize", new DoubleValue(15)));
      }
      if (!Parameters.ContainsKey("C")) {
        Parameters.Add(new LookupParameter<DoubleValue>("C"));
      }
      if (!Parameters.ContainsKey("InvertComplexityAdaption")) {
        Parameters.Add(new ValueLookupParameter<BoolValue>("InvertComplexityAdaption"));
      }
    }

    public override IOperation Apply() {
      ItemArray<SymbolicExpressionTree> trees = SymbolicExpressionTreeParameter.ActualValue;
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      // always apply Parsimony pressure if overfitting has been detected
      // otherwise appliy PP only when we are currently overfitting
      if (GenerationsParameter.ActualValue != null && GenerationsParameter.ActualValue.Value >= FirstGenerationParameter.ActualValue.Value &&
           ApplyParsimonyPressureParameter.ActualValue.Value == true) {
        var lengths = from tree in trees
                      select tree.Size;
        double k = KParameter.ActualValue.Value;

        // calculate cov(f, l) and cov(l, l^k)
        OnlineCovarianceEvaluator lengthFitnessCovEvaluator = new OnlineCovarianceEvaluator();
        OnlineCovarianceEvaluator lengthAdjLengthCovEvaluator = new OnlineCovarianceEvaluator();
        OnlineMeanAndVarianceCalculator lengthMeanCalculator = new OnlineMeanAndVarianceCalculator();
        OnlineMeanAndVarianceCalculator fitnessMeanCalculator = new OnlineMeanAndVarianceCalculator();
        OnlineMeanAndVarianceCalculator adjLengthMeanCalculator = new OnlineMeanAndVarianceCalculator();
        var lengthEnumerator = lengths.GetEnumerator();
        var qualityEnumerator = qualities.GetEnumerator();
        while (lengthEnumerator.MoveNext() & qualityEnumerator.MoveNext()) {
          double fitness = qualityEnumerator.Current.Value;
          if (!MaximizationParameter.ActualValue.Value) {
            // use f = 1 / (1 + quality) for minimization problems
            fitness = 1.0 / (1.0 + fitness);
          }
          lengthFitnessCovEvaluator.Add(lengthEnumerator.Current, fitness);
          lengthAdjLengthCovEvaluator.Add(lengthEnumerator.Current, Math.Pow(lengthEnumerator.Current, k));
          lengthMeanCalculator.Add(lengthEnumerator.Current);
          fitnessMeanCalculator.Add(fitness);
          adjLengthMeanCalculator.Add(Math.Pow(lengthEnumerator.Current, k));
        }

        //double sizeAdaption = lengthMeanCalculator.Mean * ComplexityAdaptionParameter.ActualValue.Value;
        double sizeAdaption = 100.0 * ComplexityAdaptionParameter.ActualValue.Value;
        if (InvertComplexityAdaptionParameter.ActualValue != null && InvertComplexityAdaptionParameter.ActualValue.Value) {
          sizeAdaption = -sizeAdaption;
        }
        if (lengthMeanCalculator.Mean + sizeAdaption < MinAverageSizeParameter.ActualValue.Value)
          sizeAdaption = MinAverageSizeParameter.ActualValue.Value - lengthMeanCalculator.Mean;

        //            cov(l, f) - (g(t+1) - mu(t)) avgF
        // c(t) =  --------------------------------------------
        //           cov(l, l^k) - (g(t+1) - mu(t)) E[l^k]
        double c = lengthFitnessCovEvaluator.Covariance - sizeAdaption * fitnessMeanCalculator.Mean;
        c /= lengthAdjLengthCovEvaluator.Covariance - sizeAdaption * adjLengthMeanCalculator.Mean;

        CParameter.ActualValue = new DoubleValue(c);

        // adjust fitness
        bool maximization = MaximizationParameter.ActualValue.Value;

        lengthEnumerator = lengths.GetEnumerator();
        qualityEnumerator = qualities.GetEnumerator();
        int i = 0;
        ItemArray<DoubleValue> adjQualities = new ItemArray<DoubleValue>(qualities.Length);

        while (lengthEnumerator.MoveNext() & qualityEnumerator.MoveNext()) {
          adjQualities[i++] = new DoubleValue(qualityEnumerator.Current.Value - c * Math.Pow(lengthEnumerator.Current, k));
        }
        AdjustedQualityParameter.ActualValue = adjQualities;
        double[] lengthArr = lengths.Select(x => (double)x).ToArray<double>();

        double[] adjFitess = (from f in AdjustedQualityParameter.ActualValue
                              select f.Value).ToArray<double>();
        double[] fitnessArr = (from f in QualityParameter.ActualValue
                               let normFit = maximization ? f.Value : 1.0 / (1.0 + f.Value)
                               select normFit).ToArray<double>();

        LengthCorrelationParameter.ActualValue = new DoubleValue(alglib.spearmancorr2(lengthArr, adjFitess, lengthArr.Length));
        FitnessCorrelationParameter.ActualValue = new DoubleValue(alglib.spearmancorr2(fitnessArr, adjFitess, lengthArr.Length));

      } else {
        CParameter.ActualValue = new DoubleValue(0.0);
        // adjusted fitness is equal to fitness
        AdjustedQualityParameter.ActualValue = (ItemArray<DoubleValue>)QualityParameter.ActualValue.Clone();
        FitnessCorrelationParameter.ActualValue = new DoubleValue(1.0);

        double[] lengths = (from tree in trees
                            select (double)tree.Size).ToArray<double>();

        double[] fitess = (from f in AdjustedQualityParameter.ActualValue
                           select f.Value).ToArray<double>();

        LengthCorrelationParameter.ActualValue = new DoubleValue(alglib.spearmancorr2(lengths, fitess, lengths.Length));
      }
      return base.Apply();
    }
  }
}
